"""Support for MultiIndex type of Pandas
"""
import operator
import numba
import pandas as pd
from numba.core import cgutils, types
from numba.extending import NativeValue, box, intrinsic, lower_builtin, make_attribute_wrapper, models, overload, register_model, typeof_impl, unbox
from bodo.utils.conversion import ensure_contig_if_np
from bodo.utils.typing import BodoError, check_unsupported_args, dtype_to_array_type, get_val_type_maybe_str_literal, is_overload_none


class MultiIndexType(types.Type):

    def __init__(self, array_types, names_typ=None, name_typ=None):
        names_typ = (types.none,) * len(array_types
            ) if names_typ is None else names_typ
        name_typ = types.none if name_typ is None else name_typ
        self.array_types = array_types
        self.names_typ = names_typ
        self.name_typ = name_typ
        super(MultiIndexType, self).__init__(name=
            'MultiIndexType({}, {}, {})'.format(array_types, names_typ,
            name_typ))
    ndim = 1

    def copy(self):
        return MultiIndexType(self.array_types, self.names_typ, self.name_typ)

    @property
    def nlevels(self):
        return len(self.array_types)

    @property
    def mangling_args(self):
        return self.__class__.__name__, (self._code,)


@register_model(MultiIndexType)
class MultiIndexModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        wefih__zlr = [('data', types.Tuple(fe_type.array_types)), ('names',
            types.Tuple(fe_type.names_typ)), ('name', fe_type.name_typ)]
        super(MultiIndexModel, self).__init__(dmm, fe_type, wefih__zlr)


make_attribute_wrapper(MultiIndexType, 'data', '_data')
make_attribute_wrapper(MultiIndexType, 'names', '_names')
make_attribute_wrapper(MultiIndexType, 'name', '_name')


@typeof_impl.register(pd.MultiIndex)
def typeof_multi_index(val, c):
    array_types = tuple(numba.typeof(val.levels[cvoul__xlrck].values) for
        cvoul__xlrck in range(val.nlevels))
    return MultiIndexType(array_types, tuple(get_val_type_maybe_str_literal
        (ayff__ahrtg) for ayff__ahrtg in val.names), numba.typeof(val.name))


@box(MultiIndexType)
def box_multi_index(typ, val, c):
    ekmo__ccc = c.context.insert_const_string(c.builder.module, 'pandas')
    csp__juofk = c.pyapi.import_module_noblock(ekmo__ccc)
    bff__jwm = c.pyapi.object_getattr_string(csp__juofk, 'MultiIndex')
    ykc__pczvx = cgutils.create_struct_proxy(typ)(c.context, c.builder, val)
    c.context.nrt.incref(c.builder, types.Tuple(typ.array_types),
        ykc__pczvx.data)
    gqdmc__nocb = c.pyapi.from_native_value(types.Tuple(typ.array_types),
        ykc__pczvx.data, c.env_manager)
    c.context.nrt.incref(c.builder, types.Tuple(typ.names_typ), ykc__pczvx.
        names)
    zka__ldt = c.pyapi.from_native_value(types.Tuple(typ.names_typ),
        ykc__pczvx.names, c.env_manager)
    c.context.nrt.incref(c.builder, typ.name_typ, ykc__pczvx.name)
    lknia__ypom = c.pyapi.from_native_value(typ.name_typ, ykc__pczvx.name,
        c.env_manager)
    wxml__jelks = c.pyapi.borrow_none()
    jwmyu__jlmo = c.pyapi.call_method(bff__jwm, 'from_arrays', (gqdmc__nocb,
        wxml__jelks, zka__ldt))
    c.pyapi.object_setattr_string(jwmyu__jlmo, 'name', lknia__ypom)
    c.pyapi.decref(gqdmc__nocb)
    c.pyapi.decref(zka__ldt)
    c.pyapi.decref(lknia__ypom)
    c.pyapi.decref(csp__juofk)
    c.pyapi.decref(bff__jwm)
    c.context.nrt.decref(c.builder, typ, val)
    return jwmyu__jlmo


@unbox(MultiIndexType)
def unbox_multi_index(typ, val, c):
    yht__cqu = []
    rqp__uenz = []
    for cvoul__xlrck in range(typ.nlevels):
        mad__wiff = c.pyapi.unserialize(c.pyapi.serialize_object(cvoul__xlrck))
        hqjnr__hwe = c.pyapi.call_method(val, 'get_level_values', (mad__wiff,))
        zvrl__jwaro = c.pyapi.object_getattr_string(hqjnr__hwe, 'values')
        c.pyapi.decref(hqjnr__hwe)
        c.pyapi.decref(mad__wiff)
        pwmm__dha = c.pyapi.to_native_value(typ.array_types[cvoul__xlrck],
            zvrl__jwaro).value
        yht__cqu.append(pwmm__dha)
        rqp__uenz.append(zvrl__jwaro)
    if isinstance(types.Tuple(typ.array_types), types.UniTuple):
        data = cgutils.pack_array(c.builder, yht__cqu)
    else:
        data = cgutils.pack_struct(c.builder, yht__cqu)
    zka__ldt = c.pyapi.object_getattr_string(val, 'names')
    viziy__qosmj = c.pyapi.unserialize(c.pyapi.serialize_object(tuple))
    fsal__vbrht = c.pyapi.call_function_objargs(viziy__qosmj, (zka__ldt,))
    names = c.pyapi.to_native_value(types.Tuple(typ.names_typ), fsal__vbrht
        ).value
    lknia__ypom = c.pyapi.object_getattr_string(val, 'name')
    name = c.pyapi.to_native_value(typ.name_typ, lknia__ypom).value
    ykc__pczvx = cgutils.create_struct_proxy(typ)(c.context, c.builder)
    ykc__pczvx.data = data
    ykc__pczvx.names = names
    ykc__pczvx.name = name
    for zvrl__jwaro in rqp__uenz:
        c.pyapi.decref(zvrl__jwaro)
    c.pyapi.decref(zka__ldt)
    c.pyapi.decref(viziy__qosmj)
    c.pyapi.decref(fsal__vbrht)
    c.pyapi.decref(lknia__ypom)
    return NativeValue(ykc__pczvx._getvalue())


def from_product_error_checking(iterables, sortorder, names):
    cve__grdbw = 'pandas.MultiIndex.from_product'
    mogn__ewu = dict(sortorder=sortorder)
    dwhrp__xhn = dict(sortorder=None)
    check_unsupported_args(cve__grdbw, mogn__ewu, dwhrp__xhn, package_name=
        'pandas', module_name='Index')
    if not (is_overload_none(names) or isinstance(names, types.BaseTuple)):
        raise BodoError(f'{cve__grdbw}: names must be None or a tuple.')
    elif not isinstance(iterables, types.BaseTuple):
        raise BodoError(f'{cve__grdbw}: iterables must be a tuple.')
    elif not is_overload_none(names) and len(iterables) != len(names):
        raise BodoError(
            f'{cve__grdbw}: iterables and names must be of the same length.')


def from_product(iterable, sortorder=None, names=None):
    pass


@overload(from_product)
def from_product_overload(iterables, sortorder=None, names=None):
    from_product_error_checking(iterables, sortorder, names)
    array_types = tuple(dtype_to_array_type(iterable.dtype) for iterable in
        iterables)
    if is_overload_none(names):
        names_typ = tuple([types.none] * len(iterables))
    else:
        names_typ = names.types
    hejbp__pwr = MultiIndexType(array_types, names_typ)
    ayum__evl = f'from_product_multiindex{numba.core.ir_utils.next_label()}'
    setattr(types, ayum__evl, hejbp__pwr)
    lxqg__vgki = f"""
def impl(iterables, sortorder=None, names=None):
    with numba.objmode(mi='{ayum__evl}'):
        mi = pd.MultiIndex.from_product(iterables, names=names)
    return mi
"""
    yow__hciaw = {}
    exec(lxqg__vgki, globals(), yow__hciaw)
    cqx__sht = yow__hciaw['impl']
    return cqx__sht


@intrinsic
def init_multi_index(typingctx, data, names, name=None):
    name = types.none if name is None else name
    names = types.Tuple(names.types)

    def codegen(context, builder, signature, args):
        wvv__xtvvq, qvp__djg, ubtba__tgu = args
        zyq__udt = cgutils.create_struct_proxy(signature.return_type)(context,
            builder)
        zyq__udt.data = wvv__xtvvq
        zyq__udt.names = qvp__djg
        zyq__udt.name = ubtba__tgu
        context.nrt.incref(builder, signature.args[0], wvv__xtvvq)
        context.nrt.incref(builder, signature.args[1], qvp__djg)
        context.nrt.incref(builder, signature.args[2], ubtba__tgu)
        return zyq__udt._getvalue()
    qne__nuox = MultiIndexType(data.types, names.types, name)
    return qne__nuox(data, names, name), codegen


@overload(len, no_unliteral=True)
def overload_len_pd_multiindex(A):
    if isinstance(A, MultiIndexType):
        return lambda A: len(A._data[0])


@overload(operator.getitem, no_unliteral=True)
def overload_multi_index_getitem(I, ind):
    if not isinstance(I, MultiIndexType):
        return
    if not isinstance(ind, types.Integer):
        drg__uctg = len(I.array_types)
        lxqg__vgki = 'def impl(I, ind):\n'
        lxqg__vgki += '  data = I._data\n'
        lxqg__vgki += ('  return init_multi_index(({},), I._names, I._name)\n'
            .format(', '.join(
            f'ensure_contig_if_np(data[{cvoul__xlrck}][ind])' for
            cvoul__xlrck in range(drg__uctg))))
        yow__hciaw = {}
        exec(lxqg__vgki, {'init_multi_index': init_multi_index,
            'ensure_contig_if_np': ensure_contig_if_np}, yow__hciaw)
        cqx__sht = yow__hciaw['impl']
        return cqx__sht


@lower_builtin(operator.is_, MultiIndexType, MultiIndexType)
def multi_index_is(context, builder, sig, args):
    nnode__siay, xafu__obwz = sig.args
    if nnode__siay != xafu__obwz:
        return cgutils.false_bit

    def index_is_impl(a, b):
        return (a._data is b._data and a._names is b._names and a._name is
            b._name)
    return context.compile_internal(builder, index_is_impl, sig, args)
