"""
Support for Series.str methods
"""
import operator
import re
import numba
import numpy as np
from numba.core import cgutils, types
from numba.extending import intrinsic, make_attribute_wrapper, models, overload, overload_attribute, overload_method, register_model
import bodo
from bodo.hiframes.pd_dataframe_ext import DataFrameType
from bodo.hiframes.pd_index_ext import StringIndexType
from bodo.hiframes.pd_series_ext import SeriesType
from bodo.hiframes.split_impl import get_split_view_data_ptr, get_split_view_index, string_array_split_view_type
from bodo.libs.array import get_search_regex
from bodo.libs.array_item_arr_ext import ArrayItemArrayType
from bodo.libs.str_arr_ext import get_utf8_size, pre_alloc_string_array, string_array_type
from bodo.libs.str_ext import str_findall_count
from bodo.utils.typing import BodoError, create_unsupported_overload, get_overload_const_int, get_overload_const_list, get_overload_const_str, get_overload_const_str_len, is_list_like_index_type, is_overload_constant_bool, is_overload_constant_int, is_overload_constant_list, is_overload_constant_str, is_overload_false, is_overload_none, is_overload_true, is_str_arr_type, raise_bodo_error


class SeriesStrMethodType(types.Type):

    def __init__(self, stype):
        self.stype = stype
        oxnm__xuab = 'SeriesStrMethodType({})'.format(stype)
        super(SeriesStrMethodType, self).__init__(oxnm__xuab)


@register_model(SeriesStrMethodType)
class SeriesStrModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        tdagp__edt = [('obj', fe_type.stype)]
        super(SeriesStrModel, self).__init__(dmm, fe_type, tdagp__edt)


make_attribute_wrapper(SeriesStrMethodType, 'obj', '_obj')


@intrinsic
def init_series_str_method(typingctx, obj=None):

    def codegen(context, builder, signature, args):
        uobcj__kuyjb, = args
        fxykz__jtw = signature.return_type
        ldhb__uoju = cgutils.create_struct_proxy(fxykz__jtw)(context, builder)
        ldhb__uoju.obj = uobcj__kuyjb
        context.nrt.incref(builder, signature.args[0], uobcj__kuyjb)
        return ldhb__uoju._getvalue()
    return SeriesStrMethodType(obj)(obj), codegen


def str_arg_check(func_name, arg_name, arg):
    if not isinstance(arg, types.UnicodeType) and not is_overload_constant_str(
        arg):
        raise_bodo_error(
            "Series.str.{}(): parameter '{}' expected a string object, not {}"
            .format(func_name, arg_name, arg))


def int_arg_check(func_name, arg_name, arg):
    if not isinstance(arg, types.Integer) and not is_overload_constant_int(arg
        ):
        raise BodoError(
            "Series.str.{}(): parameter '{}' expected an int object, not {}"
            .format(func_name, arg_name, arg))


def not_supported_arg_check(func_name, arg_name, arg, defval):
    if arg_name == 'na':
        if not isinstance(arg, types.Omitted) and (not isinstance(arg,
            float) or not np.isnan(arg)):
            raise BodoError(
                "Series.str.{}(): parameter '{}' is not supported, default: np.nan"
                .format(func_name, arg_name))
    elif not isinstance(arg, types.Omitted) and arg != defval:
        raise BodoError(
            "Series.str.{}(): parameter '{}' is not supported, default: {}"
            .format(func_name, arg_name, defval))


def common_validate_padding(func_name, width, fillchar):
    if is_overload_constant_str(fillchar):
        if get_overload_const_str_len(fillchar) != 1:
            raise BodoError(
                'Series.str.{}(): fillchar must be a character, not str'.
                format(func_name))
    elif not isinstance(fillchar, types.UnicodeType):
        raise BodoError('Series.str.{}(): fillchar must be a character, not {}'
            .format(func_name, fillchar))
    int_arg_check(func_name, 'width', width)


@overload_attribute(SeriesType, 'str')
def overload_series_str(S):
    if not (is_str_arr_type(S.data) or S.data ==
        string_array_split_view_type or isinstance(S.data, ArrayItemArrayType)
        ):
        raise_bodo_error(
            'Series.str: input should be a series of string or arrays')
    return lambda S: bodo.hiframes.series_str_impl.init_series_str_method(S)


@overload_method(SeriesStrMethodType, 'len', inline='always', no_unliteral=True
    )
def overload_str_method_len(S_str):

    def impl(S_str):
        S = S_str._obj
        wjtdr__eibd = bodo.hiframes.pd_series_ext.get_series_data(S)
        zbk__dfl = bodo.hiframes.pd_series_ext.get_series_index(S)
        oxnm__xuab = bodo.hiframes.pd_series_ext.get_series_name(S)
        out_arr = bodo.libs.array_kernels.get_arr_lens(wjtdr__eibd, False)
        return bodo.hiframes.pd_series_ext.init_series(out_arr, zbk__dfl,
            oxnm__xuab)
    return impl


@overload_method(SeriesStrMethodType, 'split', inline='always',
    no_unliteral=True)
def overload_str_method_split(S_str, pat=None, n=-1, expand=False):
    if not is_overload_none(pat):
        str_arg_check('split', 'pat', pat)
    int_arg_check('split', 'n', n)
    not_supported_arg_check('split', 'expand', expand, False)
    if is_overload_constant_str(pat) and len(get_overload_const_str(pat)
        ) == 1 and get_overload_const_str(pat).isascii(
        ) and is_overload_constant_int(n) and get_overload_const_int(n
        ) == -1 and S_str.stype.data == string_array_type:

        def _str_split_view_impl(S_str, pat=None, n=-1, expand=False):
            S = S_str._obj
            wjtdr__eibd = bodo.hiframes.pd_series_ext.get_series_data(S)
            zbk__dfl = bodo.hiframes.pd_series_ext.get_series_index(S)
            oxnm__xuab = bodo.hiframes.pd_series_ext.get_series_name(S)
            out_arr = bodo.hiframes.split_impl.compute_split_view(wjtdr__eibd,
                pat)
            return bodo.hiframes.pd_series_ext.init_series(out_arr,
                zbk__dfl, oxnm__xuab)
        return _str_split_view_impl

    def _str_split_impl(S_str, pat=None, n=-1, expand=False):
        S = S_str._obj
        wjtdr__eibd = bodo.hiframes.pd_series_ext.get_series_data(S)
        zbk__dfl = bodo.hiframes.pd_series_ext.get_series_index(S)
        oxnm__xuab = bodo.hiframes.pd_series_ext.get_series_name(S)
        out_arr = bodo.libs.str_ext.str_split(wjtdr__eibd, pat, n)
        return bodo.hiframes.pd_series_ext.init_series(out_arr, zbk__dfl,
            oxnm__xuab)
    return _str_split_impl


@overload_method(SeriesStrMethodType, 'get', no_unliteral=True)
def overload_str_method_get(S_str, i):
    noe__xuehq = S_str.stype.data
    if (noe__xuehq != string_array_split_view_type and not is_str_arr_type(
        noe__xuehq)) and not isinstance(noe__xuehq, ArrayItemArrayType):
        raise_bodo_error(
            'Series.str.get(): only supports input type of Series(array(item)) and Series(str)'
            )
    int_arg_check('get', 'i', i)
    if isinstance(noe__xuehq, ArrayItemArrayType):

        def _str_get_array_impl(S_str, i):
            S = S_str._obj
            wjtdr__eibd = bodo.hiframes.pd_series_ext.get_series_data(S)
            zbk__dfl = bodo.hiframes.pd_series_ext.get_series_index(S)
            oxnm__xuab = bodo.hiframes.pd_series_ext.get_series_name(S)
            out_arr = bodo.libs.array_kernels.get(wjtdr__eibd, i)
            return bodo.hiframes.pd_series_ext.init_series(out_arr,
                zbk__dfl, oxnm__xuab)
        return _str_get_array_impl
    if noe__xuehq == string_array_split_view_type:

        def _str_get_split_impl(S_str, i):
            S = S_str._obj
            wjtdr__eibd = bodo.hiframes.pd_series_ext.get_series_data(S)
            zbk__dfl = bodo.hiframes.pd_series_ext.get_series_index(S)
            oxnm__xuab = bodo.hiframes.pd_series_ext.get_series_name(S)
            numba.parfors.parfor.init_prange()
            n = len(wjtdr__eibd)
            bbgdv__iqsmm = 0
            for moxfx__afnq in numba.parfors.parfor.internal_prange(n):
                ttx__ikq, ttx__ikq, gjwbi__mqmw = get_split_view_index(
                    wjtdr__eibd, moxfx__afnq, i)
                bbgdv__iqsmm += gjwbi__mqmw
            numba.parfors.parfor.init_prange()
            out_arr = pre_alloc_string_array(n, bbgdv__iqsmm)
            for flzh__xomie in numba.parfors.parfor.internal_prange(n):
                weqo__bjjbz, zdwby__hjci, gjwbi__mqmw = get_split_view_index(
                    wjtdr__eibd, flzh__xomie, i)
                if weqo__bjjbz == 0:
                    bodo.libs.array_kernels.setna(out_arr, flzh__xomie)
                    bjtoc__cms = get_split_view_data_ptr(wjtdr__eibd, 0)
                else:
                    bodo.libs.str_arr_ext.str_arr_set_not_na(out_arr,
                        flzh__xomie)
                    bjtoc__cms = get_split_view_data_ptr(wjtdr__eibd,
                        zdwby__hjci)
                bodo.libs.str_arr_ext.setitem_str_arr_ptr(out_arr,
                    flzh__xomie, bjtoc__cms, gjwbi__mqmw)
            return bodo.hiframes.pd_series_ext.init_series(out_arr,
                zbk__dfl, oxnm__xuab)
        return _str_get_split_impl

    def _str_get_impl(S_str, i):
        S = S_str._obj
        wjtdr__eibd = bodo.hiframes.pd_series_ext.get_series_data(S)
        zbk__dfl = bodo.hiframes.pd_series_ext.get_series_index(S)
        oxnm__xuab = bodo.hiframes.pd_series_ext.get_series_name(S)
        numba.parfors.parfor.init_prange()
        n = len(wjtdr__eibd)
        numba.parfors.parfor.init_prange()
        out_arr = pre_alloc_string_array(n, -1)
        for flzh__xomie in numba.parfors.parfor.internal_prange(n):
            if bodo.libs.array_kernels.isna(wjtdr__eibd, flzh__xomie
                ) or not len(wjtdr__eibd[flzh__xomie]) > i >= -len(wjtdr__eibd
                [flzh__xomie]):
                out_arr[flzh__xomie] = ''
                bodo.libs.array_kernels.setna(out_arr, flzh__xomie)
            else:
                out_arr[flzh__xomie] = wjtdr__eibd[flzh__xomie][i]
        return bodo.hiframes.pd_series_ext.init_series(out_arr, zbk__dfl,
            oxnm__xuab)
    return _str_get_impl


@overload_method(SeriesStrMethodType, 'join', inline='always', no_unliteral
    =True)
def overload_str_method_join(S_str, sep):
    noe__xuehq = S_str.stype.data
    if (noe__xuehq != string_array_split_view_type and noe__xuehq !=
        ArrayItemArrayType(string_array_type) and not is_str_arr_type(
        noe__xuehq)):
        raise_bodo_error(
            'Series.str.join(): only supports input type of Series(list(str)) and Series(str)'
            )
    str_arg_check('join', 'sep', sep)

    def impl(S_str, sep):
        S = S_str._obj
        yotl__ckky = bodo.hiframes.pd_series_ext.get_series_data(S)
        oxnm__xuab = bodo.hiframes.pd_series_ext.get_series_name(S)
        zbk__dfl = bodo.hiframes.pd_series_ext.get_series_index(S)
        numba.parfors.parfor.init_prange()
        n = len(yotl__ckky)
        out_arr = bodo.libs.str_arr_ext.pre_alloc_string_array(n, -1)
        for flzh__xomie in numba.parfors.parfor.internal_prange(n):
            if bodo.libs.array_kernels.isna(yotl__ckky, flzh__xomie):
                out_arr[flzh__xomie] = ''
                bodo.libs.array_kernels.setna(out_arr, flzh__xomie)
            else:
                aonad__opz = yotl__ckky[flzh__xomie]
                out_arr[flzh__xomie] = sep.join(aonad__opz)
        return bodo.hiframes.pd_series_ext.init_series(out_arr, zbk__dfl,
            oxnm__xuab)
    return impl


@overload_method(SeriesStrMethodType, 'replace', inline='always',
    no_unliteral=True)
def overload_str_method_replace(S_str, pat, repl, n=-1, case=None, flags=0,
    regex=True):
    not_supported_arg_check('replace', 'n', n, -1)
    not_supported_arg_check('replace', 'case', case, None)
    str_arg_check('replace', 'pat', pat)
    str_arg_check('replace', 'repl', repl)
    int_arg_check('replace', 'flags', flags)
    if S_str.stype.data == bodo.dict_str_arr_type:

        def _str_replace_dict_impl(S_str, pat, repl, n=-1, case=None, flags
            =0, regex=True):
            S = S_str._obj
            wjtdr__eibd = bodo.hiframes.pd_series_ext.get_series_data(S)
            zbk__dfl = bodo.hiframes.pd_series_ext.get_series_index(S)
            oxnm__xuab = bodo.hiframes.pd_series_ext.get_series_name(S)
            out_arr = bodo.libs.dict_arr_ext.str_replace(wjtdr__eibd, pat,
                repl, flags, regex)
            return bodo.hiframes.pd_series_ext.init_series(out_arr,
                zbk__dfl, oxnm__xuab)
        return _str_replace_dict_impl
    if is_overload_true(regex):

        def _str_replace_regex_impl(S_str, pat, repl, n=-1, case=None,
            flags=0, regex=True):
            S = S_str._obj
            wjtdr__eibd = bodo.hiframes.pd_series_ext.get_series_data(S)
            zbk__dfl = bodo.hiframes.pd_series_ext.get_series_index(S)
            oxnm__xuab = bodo.hiframes.pd_series_ext.get_series_name(S)
            numba.parfors.parfor.init_prange()
            opktc__vvpl = re.compile(pat, flags)
            qnihx__muz = len(wjtdr__eibd)
            out_arr = pre_alloc_string_array(qnihx__muz, -1)
            for flzh__xomie in numba.parfors.parfor.internal_prange(qnihx__muz
                ):
                if bodo.libs.array_kernels.isna(wjtdr__eibd, flzh__xomie):
                    out_arr[flzh__xomie] = ''
                    bodo.libs.array_kernels.setna(out_arr, flzh__xomie)
                    continue
                out_arr[flzh__xomie] = opktc__vvpl.sub(repl, wjtdr__eibd[
                    flzh__xomie])
            return bodo.hiframes.pd_series_ext.init_series(out_arr,
                zbk__dfl, oxnm__xuab)
        return _str_replace_regex_impl
    if not is_overload_false(regex):
        raise BodoError('Series.str.replace(): regex argument should be bool')

    def _str_replace_noregex_impl(S_str, pat, repl, n=-1, case=None, flags=
        0, regex=True):
        S = S_str._obj
        wjtdr__eibd = bodo.hiframes.pd_series_ext.get_series_data(S)
        zbk__dfl = bodo.hiframes.pd_series_ext.get_series_index(S)
        oxnm__xuab = bodo.hiframes.pd_series_ext.get_series_name(S)
        numba.parfors.parfor.init_prange()
        qnihx__muz = len(wjtdr__eibd)
        numba.parfors.parfor.init_prange()
        out_arr = pre_alloc_string_array(qnihx__muz, -1)
        for flzh__xomie in numba.parfors.parfor.internal_prange(qnihx__muz):
            if bodo.libs.array_kernels.isna(wjtdr__eibd, flzh__xomie):
                out_arr[flzh__xomie] = ''
                bodo.libs.array_kernels.setna(out_arr, flzh__xomie)
                continue
            out_arr[flzh__xomie] = wjtdr__eibd[flzh__xomie].replace(pat, repl)
        return bodo.hiframes.pd_series_ext.init_series(out_arr, zbk__dfl,
            oxnm__xuab)
    return _str_replace_noregex_impl


@numba.njit
def series_contains_regex(S, pat, case, flags, na, regex):
    with numba.objmode(out_arr=bodo.boolean_array):
        out_arr = S.array._str_contains(pat, case, flags, na, regex)
    return out_arr


def is_regex_unsupported(pat):
    apt__tnosr = ['(?a', '(?i', '(?L', '(?m', '(?s', '(?u', '(?x', '(?#']
    if is_overload_constant_str(pat):
        if isinstance(pat, types.StringLiteral):
            pat = pat.literal_value
        return any([(lrjj__dmgdj in pat) for lrjj__dmgdj in apt__tnosr])
    else:
        return True


@overload_method(SeriesStrMethodType, 'contains', no_unliteral=True)
def overload_str_method_contains(S_str, pat, case=True, flags=0, na=np.nan,
    regex=True):
    not_supported_arg_check('contains', 'na', na, np.nan)
    str_arg_check('contains', 'pat', pat)
    int_arg_check('contains', 'flags', flags)
    if not is_overload_constant_bool(regex):
        raise BodoError(
            "Series.str.contains(): 'regex' argument should be a constant boolean"
            )
    if not is_overload_constant_bool(case):
        raise BodoError(
            "Series.str.contains(): 'case' argument should be a constant boolean"
            )
    rsy__oxsl = re.IGNORECASE.value
    gmneg__hqigr = 'def impl(\n'
    gmneg__hqigr += (
        '    S_str, pat, case=True, flags=0, na=np.nan, regex=True\n')
    gmneg__hqigr += '):\n'
    gmneg__hqigr += '  S = S_str._obj\n'
    gmneg__hqigr += '  arr = bodo.hiframes.pd_series_ext.get_series_data(S)\n'
    gmneg__hqigr += (
        '  index = bodo.hiframes.pd_series_ext.get_series_index(S)\n')
    gmneg__hqigr += '  name = bodo.hiframes.pd_series_ext.get_series_name(S)\n'
    gmneg__hqigr += '  l = len(arr)\n'
    gmneg__hqigr += '  out_arr = bodo.libs.bool_arr_ext.alloc_bool_array(l)\n'
    if is_overload_true(regex):
        if is_regex_unsupported(pat) or flags:
            gmneg__hqigr += """  out_arr = bodo.hiframes.series_str_impl.series_contains_regex(S, pat, case, flags, na, regex)
"""
        else:
            gmneg__hqigr += """  get_search_regex(arr, case, bodo.libs.str_ext.unicode_to_utf8(pat), out_arr)
"""
    else:
        gmneg__hqigr += '  numba.parfors.parfor.init_prange()\n'
        if is_overload_false(case):
            gmneg__hqigr += '  upper_pat = pat.upper()\n'
        gmneg__hqigr += '  for i in numba.parfors.parfor.internal_prange(l):\n'
        gmneg__hqigr += '      if bodo.libs.array_kernels.isna(arr, i):\n'
        gmneg__hqigr += '          bodo.libs.array_kernels.setna(out_arr, i)\n'
        gmneg__hqigr += '      else: \n'
        if is_overload_true(case):
            gmneg__hqigr += '          out_arr[i] = pat in arr[i]\n'
        else:
            gmneg__hqigr += (
                '          out_arr[i] = upper_pat in arr[i].upper()\n')
    gmneg__hqigr += (
        '  return bodo.hiframes.pd_series_ext.init_series(out_arr, index, name)\n'
        )
    gwaz__faa = {}
    exec(gmneg__hqigr, {'re': re, 'bodo': bodo, 'numba': numba, 'np': np,
        're_ignorecase_value': rsy__oxsl, 'get_search_regex':
        get_search_regex}, gwaz__faa)
    impl = gwaz__faa['impl']
    return impl


@overload_method(SeriesStrMethodType, 'cat', no_unliteral=True)
def overload_str_method_cat(S_str, others=None, sep=None, na_rep=None, join
    ='left'):
    if not isinstance(others, DataFrameType):
        raise_bodo_error(
            "Series.str.cat(): 'others' must be a DataFrame currently")
    if not is_overload_none(sep):
        str_arg_check('cat', 'sep', sep)
    if not is_overload_constant_str(join) or get_overload_const_str(join
        ) != 'left':
        raise_bodo_error("Series.str.cat(): 'join' not supported yet")
    gmneg__hqigr = (
        "def impl(S_str, others=None, sep=None, na_rep=None, join='left'):\n")
    gmneg__hqigr += '  S = S_str._obj\n'
    gmneg__hqigr += '  arr = bodo.hiframes.pd_series_ext.get_series_data(S)\n'
    gmneg__hqigr += (
        '  index = bodo.hiframes.pd_series_ext.get_series_index(S)\n')
    gmneg__hqigr += '  name = bodo.hiframes.pd_series_ext.get_series_name(S)\n'
    gmneg__hqigr += '  l = len(arr)\n'
    for i in range(len(others.columns)):
        gmneg__hqigr += f"""  data{i} = bodo.hiframes.pd_dataframe_ext.get_dataframe_data(others, {i})
"""
    if S_str.stype.data == bodo.dict_str_arr_type and all(yrnoh__onpv ==
        bodo.dict_str_arr_type for yrnoh__onpv in others.data):
        upe__med = ', '.join(f'data{i}' for i in range(len(others.columns)))
        gmneg__hqigr += (
            f'  out_arr = bodo.libs.dict_arr_ext.cat_dict_str((arr, {upe__med}), sep)\n'
            )
    else:
        wpky__jmeom = ' or '.join(['bodo.libs.array_kernels.isna(arr, i)'] +
            [f'bodo.libs.array_kernels.isna(data{i}, i)' for i in range(len
            (others.columns))])
        gmneg__hqigr += (
            '  out_arr = bodo.libs.str_arr_ext.pre_alloc_string_array(l, -1)\n'
            )
        gmneg__hqigr += '  numba.parfors.parfor.init_prange()\n'
        gmneg__hqigr += '  for i in numba.parfors.parfor.internal_prange(l):\n'
        gmneg__hqigr += f'      if {wpky__jmeom}:\n'
        gmneg__hqigr += '          bodo.libs.array_kernels.setna(out_arr, i)\n'
        gmneg__hqigr += '          continue\n'
        njeo__byzy = ', '.join(['arr[i]'] + [f'data{i}[i]' for i in range(
            len(others.columns))])
        xddg__jvmdb = "''" if is_overload_none(sep) else 'sep'
        gmneg__hqigr += (
            f'      out_arr[i] = {xddg__jvmdb}.join([{njeo__byzy}])\n')
    gmneg__hqigr += (
        '  return bodo.hiframes.pd_series_ext.init_series(out_arr, index, name)\n'
        )
    gwaz__faa = {}
    exec(gmneg__hqigr, {'bodo': bodo, 'numba': numba}, gwaz__faa)
    impl = gwaz__faa['impl']
    return impl


@overload_method(SeriesStrMethodType, 'count', inline='always',
    no_unliteral=True)
def overload_str_method_count(S_str, pat, flags=0):
    str_arg_check('count', 'pat', pat)
    int_arg_check('count', 'flags', flags)

    def impl(S_str, pat, flags=0):
        S = S_str._obj
        yotl__ckky = bodo.hiframes.pd_series_ext.get_series_data(S)
        oxnm__xuab = bodo.hiframes.pd_series_ext.get_series_name(S)
        zbk__dfl = bodo.hiframes.pd_series_ext.get_series_index(S)
        opktc__vvpl = re.compile(pat, flags)
        numba.parfors.parfor.init_prange()
        qnihx__muz = len(yotl__ckky)
        out_arr = bodo.libs.int_arr_ext.alloc_int_array(qnihx__muz, np.int64)
        for i in numba.parfors.parfor.internal_prange(qnihx__muz):
            if bodo.libs.array_kernels.isna(yotl__ckky, i):
                bodo.libs.array_kernels.setna(out_arr, i)
            else:
                out_arr[i] = str_findall_count(opktc__vvpl, yotl__ckky[i])
        return bodo.hiframes.pd_series_ext.init_series(out_arr, zbk__dfl,
            oxnm__xuab)
    return impl


@overload_method(SeriesStrMethodType, 'find', inline='always', no_unliteral
    =True)
def overload_str_method_find(S_str, sub, start=0, end=None):
    str_arg_check('find', 'sub', sub)
    int_arg_check('find', 'start', start)
    if not is_overload_none(end):
        int_arg_check('find', 'end', end)

    def impl(S_str, sub, start=0, end=None):
        S = S_str._obj
        yotl__ckky = bodo.hiframes.pd_series_ext.get_series_data(S)
        oxnm__xuab = bodo.hiframes.pd_series_ext.get_series_name(S)
        zbk__dfl = bodo.hiframes.pd_series_ext.get_series_index(S)
        numba.parfors.parfor.init_prange()
        qnihx__muz = len(yotl__ckky)
        out_arr = bodo.libs.int_arr_ext.alloc_int_array(qnihx__muz, np.int64)
        for i in numba.parfors.parfor.internal_prange(qnihx__muz):
            if bodo.libs.array_kernels.isna(yotl__ckky, i):
                bodo.libs.array_kernels.setna(out_arr, i)
            else:
                out_arr[i] = yotl__ckky[i].find(sub, start, end)
        return bodo.hiframes.pd_series_ext.init_series(out_arr, zbk__dfl,
            oxnm__xuab)
    return impl


@overload_method(SeriesStrMethodType, 'rfind', inline='always',
    no_unliteral=True)
def overload_str_method_rfind(S_str, sub, start=0, end=None):
    str_arg_check('rfind', 'sub', sub)
    if start != 0:
        int_arg_check('rfind', 'start', start)
    if not is_overload_none(end):
        int_arg_check('rfind', 'end', end)

    def impl(S_str, sub, start=0, end=None):
        S = S_str._obj
        yotl__ckky = bodo.hiframes.pd_series_ext.get_series_data(S)
        oxnm__xuab = bodo.hiframes.pd_series_ext.get_series_name(S)
        zbk__dfl = bodo.hiframes.pd_series_ext.get_series_index(S)
        numba.parfors.parfor.init_prange()
        qnihx__muz = len(yotl__ckky)
        out_arr = bodo.libs.int_arr_ext.alloc_int_array(qnihx__muz, np.int64)
        for i in numba.parfors.parfor.internal_prange(qnihx__muz):
            if bodo.libs.array_kernels.isna(yotl__ckky, i):
                bodo.libs.array_kernels.setna(out_arr, i)
            else:
                out_arr[i] = yotl__ckky[i].rfind(sub, start, end)
        return bodo.hiframes.pd_series_ext.init_series(out_arr, zbk__dfl,
            oxnm__xuab)
    return impl


@overload_method(SeriesStrMethodType, 'center', inline='always',
    no_unliteral=True)
def overload_str_method_center(S_str, width, fillchar=' '):
    common_validate_padding('center', width, fillchar)

    def impl(S_str, width, fillchar=' '):
        S = S_str._obj
        yotl__ckky = bodo.hiframes.pd_series_ext.get_series_data(S)
        oxnm__xuab = bodo.hiframes.pd_series_ext.get_series_name(S)
        zbk__dfl = bodo.hiframes.pd_series_ext.get_series_index(S)
        numba.parfors.parfor.init_prange()
        qnihx__muz = len(yotl__ckky)
        out_arr = bodo.libs.str_arr_ext.pre_alloc_string_array(qnihx__muz, -1)
        for flzh__xomie in numba.parfors.parfor.internal_prange(qnihx__muz):
            if bodo.libs.array_kernels.isna(yotl__ckky, flzh__xomie):
                out_arr[flzh__xomie] = ''
                bodo.libs.array_kernels.setna(out_arr, flzh__xomie)
            else:
                out_arr[flzh__xomie] = yotl__ckky[flzh__xomie].center(width,
                    fillchar)
        return bodo.hiframes.pd_series_ext.init_series(out_arr, zbk__dfl,
            oxnm__xuab)
    return impl


@overload_method(SeriesStrMethodType, 'slice_replace', inline='always',
    no_unliteral=True)
def overload_str_method_slice_replace(S_str, start=0, stop=None, repl=''):
    int_arg_check('slice_replace', 'start', start)
    if not is_overload_none(stop):
        int_arg_check('slice_replace', 'stop', stop)
    str_arg_check('slice_replace', 'repl', repl)

    def impl(S_str, start=0, stop=None, repl=''):
        S = S_str._obj
        yotl__ckky = bodo.hiframes.pd_series_ext.get_series_data(S)
        oxnm__xuab = bodo.hiframes.pd_series_ext.get_series_name(S)
        zbk__dfl = bodo.hiframes.pd_series_ext.get_series_index(S)
        numba.parfors.parfor.init_prange()
        qnihx__muz = len(yotl__ckky)
        out_arr = bodo.libs.str_arr_ext.pre_alloc_string_array(qnihx__muz, -1)
        for flzh__xomie in numba.parfors.parfor.internal_prange(qnihx__muz):
            if bodo.libs.array_kernels.isna(yotl__ckky, flzh__xomie):
                bodo.libs.array_kernels.setna(out_arr, flzh__xomie)
            else:
                if stop is not None:
                    xby__vgket = yotl__ckky[flzh__xomie][stop:]
                else:
                    xby__vgket = ''
                out_arr[flzh__xomie] = yotl__ckky[flzh__xomie][:start
                    ] + repl + xby__vgket
        return bodo.hiframes.pd_series_ext.init_series(out_arr, zbk__dfl,
            oxnm__xuab)
    return impl


@overload_method(SeriesStrMethodType, 'repeat', inline='always',
    no_unliteral=True)
def overload_str_method_repeat(S_str, repeats):
    if isinstance(repeats, types.Integer) or is_overload_constant_int(repeats):

        def impl(S_str, repeats):
            S = S_str._obj
            yotl__ckky = bodo.hiframes.pd_series_ext.get_series_data(S)
            oxnm__xuab = bodo.hiframes.pd_series_ext.get_series_name(S)
            zbk__dfl = bodo.hiframes.pd_series_ext.get_series_index(S)
            numba.parfors.parfor.init_prange()
            qnihx__muz = len(yotl__ckky)
            out_arr = bodo.libs.str_arr_ext.pre_alloc_string_array(qnihx__muz,
                -1)
            for flzh__xomie in numba.parfors.parfor.internal_prange(qnihx__muz
                ):
                if bodo.libs.array_kernels.isna(yotl__ckky, flzh__xomie):
                    bodo.libs.array_kernels.setna(out_arr, flzh__xomie)
                else:
                    out_arr[flzh__xomie] = yotl__ckky[flzh__xomie] * repeats
            return bodo.hiframes.pd_series_ext.init_series(out_arr,
                zbk__dfl, oxnm__xuab)
        return impl
    elif is_overload_constant_list(repeats):
        hbzm__dmsea = get_overload_const_list(repeats)
        fowi__jlq = all([isinstance(tndd__kmkk, int) for tndd__kmkk in
            hbzm__dmsea])
    elif is_list_like_index_type(repeats) and isinstance(repeats.dtype,
        types.Integer):
        fowi__jlq = True
    else:
        fowi__jlq = False
    if fowi__jlq:

        def impl(S_str, repeats):
            S = S_str._obj
            yotl__ckky = bodo.hiframes.pd_series_ext.get_series_data(S)
            oxnm__xuab = bodo.hiframes.pd_series_ext.get_series_name(S)
            zbk__dfl = bodo.hiframes.pd_series_ext.get_series_index(S)
            sdwy__euh = bodo.utils.conversion.coerce_to_array(repeats)
            numba.parfors.parfor.init_prange()
            qnihx__muz = len(yotl__ckky)
            out_arr = bodo.libs.str_arr_ext.pre_alloc_string_array(qnihx__muz,
                -1)
            for flzh__xomie in numba.parfors.parfor.internal_prange(qnihx__muz
                ):
                if bodo.libs.array_kernels.isna(yotl__ckky, flzh__xomie):
                    bodo.libs.array_kernels.setna(out_arr, flzh__xomie)
                else:
                    out_arr[flzh__xomie] = yotl__ckky[flzh__xomie] * sdwy__euh[
                        flzh__xomie]
            return bodo.hiframes.pd_series_ext.init_series(out_arr,
                zbk__dfl, oxnm__xuab)
        return impl
    else:
        raise BodoError(
            'Series.str.repeat(): repeats argument must either be an integer or a sequence of integers'
            )


@overload_method(SeriesStrMethodType, 'ljust', inline='always',
    no_unliteral=True)
def overload_str_method_ljust(S_str, width, fillchar=' '):
    common_validate_padding('ljust', width, fillchar)

    def impl(S_str, width, fillchar=' '):
        S = S_str._obj
        yotl__ckky = bodo.hiframes.pd_series_ext.get_series_data(S)
        oxnm__xuab = bodo.hiframes.pd_series_ext.get_series_name(S)
        zbk__dfl = bodo.hiframes.pd_series_ext.get_series_index(S)
        numba.parfors.parfor.init_prange()
        qnihx__muz = len(yotl__ckky)
        out_arr = bodo.libs.str_arr_ext.pre_alloc_string_array(qnihx__muz, -1)
        for flzh__xomie in numba.parfors.parfor.internal_prange(qnihx__muz):
            if bodo.libs.array_kernels.isna(yotl__ckky, flzh__xomie):
                out_arr[flzh__xomie] = ''
                bodo.libs.array_kernels.setna(out_arr, flzh__xomie)
            else:
                out_arr[flzh__xomie] = yotl__ckky[flzh__xomie].ljust(width,
                    fillchar)
        return bodo.hiframes.pd_series_ext.init_series(out_arr, zbk__dfl,
            oxnm__xuab)
    return impl


@overload_method(SeriesStrMethodType, 'rjust', inline='always',
    no_unliteral=True)
def overload_str_method_rjust(S_str, width, fillchar=' '):
    common_validate_padding('rjust', width, fillchar)

    def impl(S_str, width, fillchar=' '):
        S = S_str._obj
        yotl__ckky = bodo.hiframes.pd_series_ext.get_series_data(S)
        oxnm__xuab = bodo.hiframes.pd_series_ext.get_series_name(S)
        zbk__dfl = bodo.hiframes.pd_series_ext.get_series_index(S)
        numba.parfors.parfor.init_prange()
        qnihx__muz = len(yotl__ckky)
        out_arr = bodo.libs.str_arr_ext.pre_alloc_string_array(qnihx__muz, -1)
        for flzh__xomie in numba.parfors.parfor.internal_prange(qnihx__muz):
            if bodo.libs.array_kernels.isna(yotl__ckky, flzh__xomie):
                out_arr[flzh__xomie] = ''
                bodo.libs.array_kernels.setna(out_arr, flzh__xomie)
            else:
                out_arr[flzh__xomie] = yotl__ckky[flzh__xomie].rjust(width,
                    fillchar)
        return bodo.hiframes.pd_series_ext.init_series(out_arr, zbk__dfl,
            oxnm__xuab)
    return impl


@overload_method(SeriesStrMethodType, 'pad', no_unliteral=True)
def overload_str_method_pad(S_str, width, side='left', fillchar=' '):
    common_validate_padding('pad', width, fillchar)
    if is_overload_constant_str(side):
        if get_overload_const_str(side) not in ['left', 'right', 'both']:
            raise BodoError('Series.str.pad(): Invalid Side')
    else:
        raise BodoError('Series.str.pad(): Invalid Side')

    def impl(S_str, width, side='left', fillchar=' '):
        S = S_str._obj
        yotl__ckky = bodo.hiframes.pd_series_ext.get_series_data(S)
        oxnm__xuab = bodo.hiframes.pd_series_ext.get_series_name(S)
        zbk__dfl = bodo.hiframes.pd_series_ext.get_series_index(S)
        numba.parfors.parfor.init_prange()
        qnihx__muz = len(yotl__ckky)
        out_arr = bodo.libs.str_arr_ext.pre_alloc_string_array(qnihx__muz, -1)
        for flzh__xomie in numba.parfors.parfor.internal_prange(qnihx__muz):
            if bodo.libs.array_kernels.isna(yotl__ckky, flzh__xomie):
                out_arr[flzh__xomie] = ''
                bodo.libs.array_kernels.setna(out_arr, flzh__xomie)
            elif side == 'left':
                out_arr[flzh__xomie] = yotl__ckky[flzh__xomie].rjust(width,
                    fillchar)
            elif side == 'right':
                out_arr[flzh__xomie] = yotl__ckky[flzh__xomie].ljust(width,
                    fillchar)
            elif side == 'both':
                out_arr[flzh__xomie] = yotl__ckky[flzh__xomie].center(width,
                    fillchar)
        return bodo.hiframes.pd_series_ext.init_series(out_arr, zbk__dfl,
            oxnm__xuab)
    return impl


@overload_method(SeriesStrMethodType, 'zfill', inline='always',
    no_unliteral=True)
def overload_str_method_zfill(S_str, width):
    int_arg_check('zfill', 'width', width)

    def impl(S_str, width):
        S = S_str._obj
        yotl__ckky = bodo.hiframes.pd_series_ext.get_series_data(S)
        oxnm__xuab = bodo.hiframes.pd_series_ext.get_series_name(S)
        zbk__dfl = bodo.hiframes.pd_series_ext.get_series_index(S)
        numba.parfors.parfor.init_prange()
        qnihx__muz = len(yotl__ckky)
        out_arr = bodo.libs.str_arr_ext.pre_alloc_string_array(qnihx__muz, -1)
        for flzh__xomie in numba.parfors.parfor.internal_prange(qnihx__muz):
            if bodo.libs.array_kernels.isna(yotl__ckky, flzh__xomie):
                out_arr[flzh__xomie] = ''
                bodo.libs.array_kernels.setna(out_arr, flzh__xomie)
            else:
                out_arr[flzh__xomie] = yotl__ckky[flzh__xomie].zfill(width)
        return bodo.hiframes.pd_series_ext.init_series(out_arr, zbk__dfl,
            oxnm__xuab)
    return impl


@overload_method(SeriesStrMethodType, 'slice', no_unliteral=True)
def overload_str_method_slice(S_str, start=None, stop=None, step=None):
    if not is_overload_none(start):
        int_arg_check('slice', 'start', start)
    if not is_overload_none(stop):
        int_arg_check('slice', 'stop', stop)
    if not is_overload_none(step):
        int_arg_check('slice', 'step', step)

    def impl(S_str, start=None, stop=None, step=None):
        S = S_str._obj
        yotl__ckky = bodo.hiframes.pd_series_ext.get_series_data(S)
        oxnm__xuab = bodo.hiframes.pd_series_ext.get_series_name(S)
        zbk__dfl = bodo.hiframes.pd_series_ext.get_series_index(S)
        numba.parfors.parfor.init_prange()
        qnihx__muz = len(yotl__ckky)
        out_arr = bodo.libs.str_arr_ext.pre_alloc_string_array(qnihx__muz, -1)
        for flzh__xomie in numba.parfors.parfor.internal_prange(qnihx__muz):
            if bodo.libs.array_kernels.isna(yotl__ckky, flzh__xomie):
                out_arr[flzh__xomie] = ''
                bodo.libs.array_kernels.setna(out_arr, flzh__xomie)
            else:
                out_arr[flzh__xomie] = yotl__ckky[flzh__xomie][start:stop:step]
        return bodo.hiframes.pd_series_ext.init_series(out_arr, zbk__dfl,
            oxnm__xuab)
    return impl


@overload_method(SeriesStrMethodType, 'startswith', inline='always',
    no_unliteral=True)
def overload_str_method_startswith(S_str, pat, na=np.nan):
    not_supported_arg_check('startswith', 'na', na, np.nan)
    str_arg_check('startswith', 'pat', pat)

    def impl(S_str, pat, na=np.nan):
        S = S_str._obj
        yotl__ckky = bodo.hiframes.pd_series_ext.get_series_data(S)
        oxnm__xuab = bodo.hiframes.pd_series_ext.get_series_name(S)
        zbk__dfl = bodo.hiframes.pd_series_ext.get_series_index(S)
        numba.parfors.parfor.init_prange()
        qnihx__muz = len(yotl__ckky)
        out_arr = bodo.libs.bool_arr_ext.alloc_bool_array(qnihx__muz)
        for i in numba.parfors.parfor.internal_prange(qnihx__muz):
            if bodo.libs.array_kernels.isna(yotl__ckky, i):
                bodo.libs.array_kernels.setna(out_arr, i)
            else:
                out_arr[i] = yotl__ckky[i].startswith(pat)
        return bodo.hiframes.pd_series_ext.init_series(out_arr, zbk__dfl,
            oxnm__xuab)
    return impl


@overload_method(SeriesStrMethodType, 'endswith', inline='always',
    no_unliteral=True)
def overload_str_method_endswith(S_str, pat, na=np.nan):
    not_supported_arg_check('endswith', 'na', na, np.nan)
    str_arg_check('endswith', 'pat', pat)

    def impl(S_str, pat, na=np.nan):
        S = S_str._obj
        yotl__ckky = bodo.hiframes.pd_series_ext.get_series_data(S)
        oxnm__xuab = bodo.hiframes.pd_series_ext.get_series_name(S)
        zbk__dfl = bodo.hiframes.pd_series_ext.get_series_index(S)
        numba.parfors.parfor.init_prange()
        qnihx__muz = len(yotl__ckky)
        out_arr = bodo.libs.bool_arr_ext.alloc_bool_array(qnihx__muz)
        for i in numba.parfors.parfor.internal_prange(qnihx__muz):
            if bodo.libs.array_kernels.isna(yotl__ckky, i):
                bodo.libs.array_kernels.setna(out_arr, i)
            else:
                out_arr[i] = yotl__ckky[i].endswith(pat)
        return bodo.hiframes.pd_series_ext.init_series(out_arr, zbk__dfl,
            oxnm__xuab)
    return impl


@overload(operator.getitem, no_unliteral=True)
def overload_str_method_getitem(S_str, ind):
    if not isinstance(S_str, SeriesStrMethodType):
        return
    if not isinstance(types.unliteral(ind), (types.SliceType, types.Integer)):
        raise BodoError(
            'index input to Series.str[] should be a slice or an integer')
    if isinstance(ind, types.SliceType):
        return lambda S_str, ind: S_str.slice(ind.start, ind.stop, ind.step)
    if isinstance(types.unliteral(ind), types.Integer):
        return lambda S_str, ind: S_str.get(ind)


@overload_method(SeriesStrMethodType, 'extract', inline='always',
    no_unliteral=True)
def overload_str_method_extract(S_str, pat, flags=0, expand=True):
    if not is_overload_constant_bool(expand):
        raise BodoError(
            "Series.str.extract(): 'expand' argument should be a constant bool"
            )
    sotkb__mkxn, regex = _get_column_names_from_regex(pat, flags, 'extract')
    zckfy__vupl = len(sotkb__mkxn)
    gmneg__hqigr = 'def impl(S_str, pat, flags=0, expand=True):\n'
    gmneg__hqigr += '  regex = re.compile(pat, flags=flags)\n'
    gmneg__hqigr += '  S = S_str._obj\n'
    gmneg__hqigr += (
        '  str_arr = bodo.hiframes.pd_series_ext.get_series_data(S)\n')
    gmneg__hqigr += (
        '  index = bodo.hiframes.pd_series_ext.get_series_index(S)\n')
    gmneg__hqigr += '  name = bodo.hiframes.pd_series_ext.get_series_name(S)\n'
    gmneg__hqigr += '  numba.parfors.parfor.init_prange()\n'
    gmneg__hqigr += '  n = len(str_arr)\n'
    for i in range(zckfy__vupl):
        gmneg__hqigr += (
            '  out_arr_{0} = bodo.libs.str_arr_ext.pre_alloc_string_array(n, -1)\n'
            .format(i))
    gmneg__hqigr += '  for j in numba.parfors.parfor.internal_prange(n):\n'
    gmneg__hqigr += '      if bodo.libs.array_kernels.isna(str_arr, j):\n'
    for i in range(zckfy__vupl):
        gmneg__hqigr += "          out_arr_{}[j] = ''\n".format(i)
        gmneg__hqigr += (
            '          bodo.libs.array_kernels.setna(out_arr_{}, j)\n'.
            format(i))
    gmneg__hqigr += '      else:\n'
    gmneg__hqigr += '          m = regex.search(str_arr[j])\n'
    gmneg__hqigr += '          if m:\n'
    gmneg__hqigr += '            g = m.groups()\n'
    for i in range(zckfy__vupl):
        gmneg__hqigr += '            out_arr_{0}[j] = g[{0}]\n'.format(i)
    gmneg__hqigr += '          else:\n'
    for i in range(zckfy__vupl):
        gmneg__hqigr += "            out_arr_{}[j] = ''\n".format(i)
        gmneg__hqigr += (
            '            bodo.libs.array_kernels.setna(out_arr_{}, j)\n'.
            format(i))
    if is_overload_false(expand) and regex.groups == 1:
        oxnm__xuab = "'{}'".format(list(regex.groupindex.keys()).pop()) if len(
            regex.groupindex.keys()) > 0 else 'name'
        gmneg__hqigr += (
            '  return bodo.hiframes.pd_series_ext.init_series(out_arr_0, index, {})\n'
            .format(oxnm__xuab))
        gwaz__faa = {}
        exec(gmneg__hqigr, {'re': re, 'bodo': bodo, 'numba': numba,
            'get_utf8_size': get_utf8_size}, gwaz__faa)
        impl = gwaz__faa['impl']
        return impl
    vla__gboq = ', '.join('out_arr_{}'.format(i) for i in range(zckfy__vupl))
    impl = bodo.hiframes.dataframe_impl._gen_init_df(gmneg__hqigr,
        sotkb__mkxn, vla__gboq, 'index', extra_globals={'get_utf8_size':
        get_utf8_size, 're': re})
    return impl


@overload_method(SeriesStrMethodType, 'extractall', inline='always',
    no_unliteral=True)
def overload_str_method_extractall(S_str, pat, flags=0):
    sotkb__mkxn, ttx__ikq = _get_column_names_from_regex(pat, flags,
        'extractall')
    zckfy__vupl = len(sotkb__mkxn)
    bqm__zop = isinstance(S_str.stype.index, StringIndexType)
    gmneg__hqigr = 'def impl(S_str, pat, flags=0):\n'
    gmneg__hqigr += '  regex = re.compile(pat, flags=flags)\n'
    gmneg__hqigr += '  S = S_str._obj\n'
    gmneg__hqigr += (
        '  str_arr = bodo.hiframes.pd_series_ext.get_series_data(S)\n')
    gmneg__hqigr += (
        '  index = bodo.hiframes.pd_series_ext.get_series_index(S)\n')
    gmneg__hqigr += '  name = bodo.hiframes.pd_series_ext.get_series_name(S)\n'
    gmneg__hqigr += (
        '  index_arr = bodo.utils.conversion.index_to_array(index)\n')
    gmneg__hqigr += (
        '  index_name = bodo.hiframes.pd_index_ext.get_index_name(index)\n')
    gmneg__hqigr += '  numba.parfors.parfor.init_prange()\n'
    gmneg__hqigr += '  n = len(str_arr)\n'
    gmneg__hqigr += '  out_n_l = [0]\n'
    for i in range(zckfy__vupl):
        gmneg__hqigr += '  num_chars_{} = 0\n'.format(i)
    if bqm__zop:
        gmneg__hqigr += '  index_num_chars = 0\n'
    gmneg__hqigr += '  for i in numba.parfors.parfor.internal_prange(n):\n'
    if bqm__zop:
        gmneg__hqigr += (
            '      index_num_chars += get_utf8_size(index_arr[i])\n')
    gmneg__hqigr += '      if bodo.libs.array_kernels.isna(str_arr, i):\n'
    gmneg__hqigr += '          continue\n'
    gmneg__hqigr += '      m = regex.findall(str_arr[i])\n'
    gmneg__hqigr += '      out_n_l[0] += len(m)\n'
    for i in range(zckfy__vupl):
        gmneg__hqigr += '      l_{} = 0\n'.format(i)
    gmneg__hqigr += '      for s in m:\n'
    for i in range(zckfy__vupl):
        gmneg__hqigr += '        l_{} += get_utf8_size(s{})\n'.format(i, 
            '[{}]'.format(i) if zckfy__vupl > 1 else '')
    for i in range(zckfy__vupl):
        gmneg__hqigr += '      num_chars_{0} += l_{0}\n'.format(i)
    gmneg__hqigr += (
        '  out_n = bodo.libs.distributed_api.local_alloc_size(out_n_l[0], str_arr)\n'
        )
    for i in range(zckfy__vupl):
        gmneg__hqigr += (
            """  out_arr_{0} = bodo.libs.str_arr_ext.pre_alloc_string_array(out_n, num_chars_{0})
"""
            .format(i))
    if bqm__zop:
        gmneg__hqigr += """  out_ind_arr = bodo.libs.str_arr_ext.pre_alloc_string_array(out_n, index_num_chars)
"""
    else:
        gmneg__hqigr += '  out_ind_arr = np.empty(out_n, index_arr.dtype)\n'
    gmneg__hqigr += '  out_match_arr = np.empty(out_n, np.int64)\n'
    gmneg__hqigr += '  out_ind = 0\n'
    gmneg__hqigr += '  for j in numba.parfors.parfor.internal_prange(n):\n'
    gmneg__hqigr += '      if bodo.libs.array_kernels.isna(str_arr, j):\n'
    gmneg__hqigr += '          continue\n'
    gmneg__hqigr += '      m = regex.findall(str_arr[j])\n'
    gmneg__hqigr += '      for k, s in enumerate(m):\n'
    for i in range(zckfy__vupl):
        gmneg__hqigr += (
            """        bodo.libs.distributed_api.set_arr_local(out_arr_{}, out_ind, s{})
"""
            .format(i, '[{}]'.format(i) if zckfy__vupl > 1 else ''))
    gmneg__hqigr += """        bodo.libs.distributed_api.set_arr_local(out_ind_arr, out_ind, index_arr[j])
"""
    gmneg__hqigr += (
        '        bodo.libs.distributed_api.set_arr_local(out_match_arr, out_ind, k)\n'
        )
    gmneg__hqigr += '        out_ind += 1\n'
    gmneg__hqigr += (
        '  out_index = bodo.hiframes.pd_multi_index_ext.init_multi_index(\n')
    gmneg__hqigr += (
        "    (out_ind_arr, out_match_arr), (index_name, 'match'))\n")
    vla__gboq = ', '.join('out_arr_{}'.format(i) for i in range(zckfy__vupl))
    impl = bodo.hiframes.dataframe_impl._gen_init_df(gmneg__hqigr,
        sotkb__mkxn, vla__gboq, 'out_index', extra_globals={'get_utf8_size':
        get_utf8_size, 're': re})
    return impl


def _get_column_names_from_regex(pat, flags, func_name):
    if not is_overload_constant_str(pat):
        raise BodoError(
            "Series.str.{}(): 'pat' argument should be a constant string".
            format(func_name))
    if not is_overload_constant_int(flags):
        raise BodoError(
            "Series.str.{}(): 'flags' argument should be a constant int".
            format(func_name))
    pat = get_overload_const_str(pat)
    flags = get_overload_const_int(flags)
    regex = re.compile(pat, flags=flags)
    if regex.groups == 0:
        raise BodoError(
            'Series.str.{}(): pattern {} contains no capture groups'.format
            (func_name, pat))
    ejlf__mug = dict(zip(regex.groupindex.values(), regex.groupindex.keys()))
    sotkb__mkxn = [ejlf__mug.get(1 + i, i) for i in range(regex.groups)]
    return sotkb__mkxn, regex


def create_str2str_methods_overload(func_name):
    if func_name in ['lstrip', 'rstrip', 'strip']:
        gmneg__hqigr = 'def f(S_str, to_strip=None):\n'
    else:
        gmneg__hqigr = 'def f(S_str):\n'
    gmneg__hqigr += '    S = S_str._obj\n'
    gmneg__hqigr += (
        '    str_arr = bodo.hiframes.pd_series_ext.get_series_data(S)\n')
    gmneg__hqigr += '    str_arr = decode_if_dict_array(str_arr)\n'
    gmneg__hqigr += (
        '    index = bodo.hiframes.pd_series_ext.get_series_index(S)\n')
    gmneg__hqigr += (
        '    name = bodo.hiframes.pd_series_ext.get_series_name(S)\n')
    gmneg__hqigr += '    numba.parfors.parfor.init_prange()\n'
    gmneg__hqigr += '    n = len(str_arr)\n'
    if func_name in ('capitalize', 'lower', 'swapcase', 'title', 'upper'):
        gmneg__hqigr += '    num_chars = num_total_chars(str_arr)\n'
    else:
        gmneg__hqigr += '    num_chars = -1\n'
    gmneg__hqigr += (
        '    out_arr = bodo.libs.str_arr_ext.pre_alloc_string_array(n, num_chars)\n'
        )
    gmneg__hqigr += '    for j in numba.parfors.parfor.internal_prange(n):\n'
    gmneg__hqigr += '        if bodo.libs.array_kernels.isna(str_arr, j):\n'
    gmneg__hqigr += '            out_arr[j] = ""\n'
    gmneg__hqigr += '            bodo.libs.array_kernels.setna(out_arr, j)\n'
    gmneg__hqigr += '        else:\n'
    if func_name in ['lstrip', 'rstrip', 'strip']:
        gmneg__hqigr += ('            out_arr[j] = str_arr[j].{}(to_strip)\n'
            .format(func_name))
    else:
        gmneg__hqigr += '            out_arr[j] = str_arr[j].{}()\n'.format(
            func_name)
    gmneg__hqigr += (
        '    return bodo.hiframes.pd_series_ext.init_series(out_arr, index, name)\n'
        )
    gwaz__faa = {}
    exec(gmneg__hqigr, {'bodo': bodo, 'numba': numba, 'num_total_chars':
        bodo.libs.str_arr_ext.num_total_chars, 'get_utf8_size': bodo.libs.
        str_arr_ext.get_utf8_size, 'decode_if_dict_array': bodo.utils.
        typing.decode_if_dict_array}, gwaz__faa)
    rplh__ejy = gwaz__faa['f']
    if func_name in ['lstrip', 'rstrip', 'strip']:

        def overload_strip_method(S_str, to_strip=None):
            if not is_overload_none(to_strip):
                str_arg_check(func_name, 'to_strip', to_strip)
            return rplh__ejy
        return overload_strip_method
    else:

        def overload_str2str_methods(S_str):
            return rplh__ejy
        return overload_str2str_methods


def create_str2bool_methods_overload(func_name):

    def overload_str2bool_methods(S_str):
        gmneg__hqigr = 'def f(S_str):\n'
        gmneg__hqigr += '    S = S_str._obj\n'
        gmneg__hqigr += (
            '    str_arr = bodo.hiframes.pd_series_ext.get_series_data(S)\n')
        gmneg__hqigr += (
            '    index = bodo.hiframes.pd_series_ext.get_series_index(S)\n')
        gmneg__hqigr += (
            '    name = bodo.hiframes.pd_series_ext.get_series_name(S)\n')
        gmneg__hqigr += '    numba.parfors.parfor.init_prange()\n'
        gmneg__hqigr += '    l = len(str_arr)\n'
        gmneg__hqigr += (
            '    out_arr = bodo.libs.bool_arr_ext.alloc_bool_array(l)\n')
        gmneg__hqigr += (
            '    for i in numba.parfors.parfor.internal_prange(l):\n')
        gmneg__hqigr += (
            '        if bodo.libs.array_kernels.isna(str_arr, i):\n')
        gmneg__hqigr += (
            '            bodo.libs.array_kernels.setna(out_arr, i)\n')
        gmneg__hqigr += '        else:\n'
        gmneg__hqigr += ('            out_arr[i] = np.bool_(str_arr[i].{}())\n'
            .format(func_name))
        gmneg__hqigr += '    return bodo.hiframes.pd_series_ext.init_series(\n'
        gmneg__hqigr += '      out_arr,index, name)\n'
        gwaz__faa = {}
        exec(gmneg__hqigr, {'bodo': bodo, 'numba': numba, 'np': np}, gwaz__faa)
        rplh__ejy = gwaz__faa['f']
        return rplh__ejy
    return overload_str2bool_methods


def _install_str2str_methods():
    for rfi__lfth in bodo.hiframes.pd_series_ext.str2str_methods:
        zycel__bbub = create_str2str_methods_overload(rfi__lfth)
        overload_method(SeriesStrMethodType, rfi__lfth, inline='always',
            no_unliteral=True)(zycel__bbub)


def _install_str2bool_methods():
    for rfi__lfth in bodo.hiframes.pd_series_ext.str2bool_methods:
        zycel__bbub = create_str2bool_methods_overload(rfi__lfth)
        overload_method(SeriesStrMethodType, rfi__lfth, inline='always',
            no_unliteral=True)(zycel__bbub)


_install_str2str_methods()
_install_str2bool_methods()


@overload_attribute(SeriesType, 'cat')
def overload_series_cat(s):
    if not isinstance(s.dtype, bodo.hiframes.pd_categorical_ext.
        PDCategoricalDtype):
        raise BodoError('Can only use .cat accessor with categorical values.')
    return lambda s: bodo.hiframes.series_str_impl.init_series_cat_method(s)


class SeriesCatMethodType(types.Type):

    def __init__(self, stype):
        self.stype = stype
        oxnm__xuab = 'SeriesCatMethodType({})'.format(stype)
        super(SeriesCatMethodType, self).__init__(oxnm__xuab)

    @property
    def mangling_args(self):
        return self.__class__.__name__, (self._code,)


@register_model(SeriesCatMethodType)
class SeriesCatModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        tdagp__edt = [('obj', fe_type.stype)]
        super(SeriesCatModel, self).__init__(dmm, fe_type, tdagp__edt)


make_attribute_wrapper(SeriesCatMethodType, 'obj', '_obj')


@intrinsic
def init_series_cat_method(typingctx, obj=None):

    def codegen(context, builder, signature, args):
        uobcj__kuyjb, = args
        wyzds__lflq = signature.return_type
        lomgw__zfmf = cgutils.create_struct_proxy(wyzds__lflq)(context, builder
            )
        lomgw__zfmf.obj = uobcj__kuyjb
        context.nrt.incref(builder, signature.args[0], uobcj__kuyjb)
        return lomgw__zfmf._getvalue()
    return SeriesCatMethodType(obj)(obj), codegen


@overload_attribute(SeriesCatMethodType, 'codes')
def series_cat_codes_overload(S_dt):

    def impl(S_dt):
        S = S_dt._obj
        wjtdr__eibd = bodo.hiframes.pd_series_ext.get_series_data(S)
        zbk__dfl = bodo.hiframes.pd_series_ext.get_series_index(S)
        oxnm__xuab = None
        return bodo.hiframes.pd_series_ext.init_series(bodo.hiframes.
            pd_categorical_ext.get_categorical_arr_codes(wjtdr__eibd),
            zbk__dfl, oxnm__xuab)
    return impl


unsupported_cat_attrs = {'categories', 'ordered'}
unsupported_cat_methods = {'rename_categories', 'reorder_categories',
    'add_categories', 'remove_categories', 'remove_unused_categories',
    'set_categories', 'as_ordered', 'as_unordered'}


def _install_catseries_unsupported():
    for pghw__ffv in unsupported_cat_attrs:
        tez__scxnl = 'Series.cat.' + pghw__ffv
        overload_attribute(SeriesCatMethodType, pghw__ffv)(
            create_unsupported_overload(tez__scxnl))
    for ejnf__jige in unsupported_cat_methods:
        tez__scxnl = 'Series.cat.' + ejnf__jige
        overload_method(SeriesCatMethodType, ejnf__jige)(
            create_unsupported_overload(tez__scxnl))


_install_catseries_unsupported()
unsupported_str_methods = {'casefold', 'decode', 'encode', 'findall',
    'fullmatch', 'index', 'match', 'normalize', 'partition', 'rindex',
    'rpartition', 'slice_replace', 'rsplit', 'translate', 'wrap', 'get_dummies'
    }


def _install_strseries_unsupported():
    for ejnf__jige in unsupported_str_methods:
        tez__scxnl = 'Series.str.' + ejnf__jige
        overload_method(SeriesStrMethodType, ejnf__jige)(
            create_unsupported_overload(tez__scxnl))


_install_strseries_unsupported()
