"""
Array of intervals corresponding to IntervalArray of Pandas.
Used for IntervalIndex, which is necessary for Series.value_counts() with 'bins'
argument.
"""
import numba
import pandas as pd
from numba.core import cgutils, types
from numba.extending import NativeValue, box, intrinsic, make_attribute_wrapper, models, overload, overload_attribute, overload_method, register_model, typeof_impl, unbox
from numba.parfors.array_analysis import ArrayAnalysis
import bodo


class IntervalType(types.Type):

    def __init__(self):
        super(IntervalType, self).__init__('IntervalType()')


class IntervalArrayType(types.ArrayCompatible):

    def __init__(self, arr_type):
        self.arr_type = arr_type
        self.dtype = IntervalType()
        super(IntervalArrayType, self).__init__(name=
            f'IntervalArrayType({arr_type})')

    @property
    def as_array(self):
        return types.Array(types.undefined, 1, 'C')

    def copy(self):
        return IntervalArrayType(self.arr_type)

    @property
    def mangling_args(self):
        return self.__class__.__name__, (self._code,)


@register_model(IntervalArrayType)
class IntervalArrayModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        zmim__zaex = [('left', fe_type.arr_type), ('right', fe_type.arr_type)]
        models.StructModel.__init__(self, dmm, fe_type, zmim__zaex)


make_attribute_wrapper(IntervalArrayType, 'left', '_left')
make_attribute_wrapper(IntervalArrayType, 'right', '_right')


@typeof_impl.register(pd.arrays.IntervalArray)
def typeof_interval_array(val, c):
    arr_type = bodo.typeof(val._left)
    return IntervalArrayType(arr_type)


@intrinsic
def init_interval_array(typingctx, left, right=None):
    assert left == right, 'Interval left/right array types should be the same'

    def codegen(context, builder, signature, args):
        koz__arwz, jryz__phti = args
        skpl__qtmb = cgutils.create_struct_proxy(signature.return_type)(context
            , builder)
        skpl__qtmb.left = koz__arwz
        skpl__qtmb.right = jryz__phti
        context.nrt.incref(builder, signature.args[0], koz__arwz)
        context.nrt.incref(builder, signature.args[1], jryz__phti)
        return skpl__qtmb._getvalue()
    xlck__cmd = IntervalArrayType(left)
    fomfs__xzr = xlck__cmd(left, right)
    return fomfs__xzr, codegen


def init_interval_array_equiv(self, scope, equiv_set, loc, args, kws):
    assert len(args) == 2 and not kws
    qwuj__kar = []
    for dwufi__ayrt in args:
        giahb__cvt = equiv_set.get_shape(dwufi__ayrt)
        if giahb__cvt is not None:
            qwuj__kar.append(giahb__cvt[0])
    if len(qwuj__kar) > 1:
        equiv_set.insert_equiv(*qwuj__kar)
    left = args[0]
    if equiv_set.has_shape(left):
        return ArrayAnalysis.AnalyzeResult(shape=left, pre=[])
    return None


(ArrayAnalysis._analyze_op_call_bodo_libs_interval_arr_ext_init_interval_array
    ) = init_interval_array_equiv


def alias_ext_init_interval_array(lhs_name, args, alias_map, arg_aliases):
    assert len(args) == 2
    numba.core.ir_utils._add_alias(lhs_name, args[0].name, alias_map,
        arg_aliases)
    numba.core.ir_utils._add_alias(lhs_name, args[1].name, alias_map,
        arg_aliases)


numba.core.ir_utils.alias_func_extensions['init_interval_array',
    'bodo.libs.int_arr_ext'] = alias_ext_init_interval_array


@box(IntervalArrayType)
def box_interval_arr(typ, val, c):
    skpl__qtmb = cgutils.create_struct_proxy(typ)(c.context, c.builder, val)
    c.context.nrt.incref(c.builder, typ.arr_type, skpl__qtmb.left)
    rnli__dcnsd = c.pyapi.from_native_value(typ.arr_type, skpl__qtmb.left,
        c.env_manager)
    c.context.nrt.incref(c.builder, typ.arr_type, skpl__qtmb.right)
    xgu__htcv = c.pyapi.from_native_value(typ.arr_type, skpl__qtmb.right, c
        .env_manager)
    jrhn__nqks = c.context.insert_const_string(c.builder.module, 'pandas')
    sybb__quzll = c.pyapi.import_module_noblock(jrhn__nqks)
    gqr__wtho = c.pyapi.object_getattr_string(sybb__quzll, 'arrays')
    xoyf__ajncf = c.pyapi.object_getattr_string(gqr__wtho, 'IntervalArray')
    foxc__qjm = c.pyapi.call_method(xoyf__ajncf, 'from_arrays', (
        rnli__dcnsd, xgu__htcv))
    c.pyapi.decref(rnli__dcnsd)
    c.pyapi.decref(xgu__htcv)
    c.pyapi.decref(sybb__quzll)
    c.pyapi.decref(gqr__wtho)
    c.pyapi.decref(xoyf__ajncf)
    c.context.nrt.decref(c.builder, typ, val)
    return foxc__qjm


@unbox(IntervalArrayType)
def unbox_interval_arr(typ, val, c):
    rnli__dcnsd = c.pyapi.object_getattr_string(val, '_left')
    left = c.pyapi.to_native_value(typ.arr_type, rnli__dcnsd).value
    c.pyapi.decref(rnli__dcnsd)
    xgu__htcv = c.pyapi.object_getattr_string(val, '_right')
    right = c.pyapi.to_native_value(typ.arr_type, xgu__htcv).value
    c.pyapi.decref(xgu__htcv)
    skpl__qtmb = cgutils.create_struct_proxy(typ)(c.context, c.builder)
    skpl__qtmb.left = left
    skpl__qtmb.right = right
    lqapr__afyyi = cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return NativeValue(skpl__qtmb._getvalue(), is_error=lqapr__afyyi)


@overload(len, no_unliteral=True)
def overload_interval_arr_len(A):
    if isinstance(A, IntervalArrayType):
        return lambda A: len(A._left)


@overload_attribute(IntervalArrayType, 'shape')
def overload_interval_arr_shape(A):
    return lambda A: (len(A._left),)


@overload_attribute(IntervalArrayType, 'ndim')
def overload_interval_arr_ndim(A):
    return lambda A: 1


@overload_attribute(IntervalArrayType, 'nbytes')
def overload_interval_arr_nbytes(A):
    return lambda A: A._left.nbytes + A._right.nbytes


@overload_method(IntervalArrayType, 'copy', no_unliteral=True)
def overload_interval_arr_copy(A):
    return lambda A: bodo.libs.interval_arr_ext.init_interval_array(A._left
        .copy(), A._right.copy())
